% ========================================================================
%   A demo script to reconstruct high-resolution QSM from SPICE data
%
%   This code has not been fully optimized, just release to reproduce the
%   QSM results for the paper:
%   Guo, R, Zhao, Y, Li, Y, et al. Simultaneous QSM and metabolic imaging 
%   of the brain using SPICE: Further improvements in data acquisition and 
%   processing. Magn Reson Med. 2020; 00: 1– 8. 
%   https://doi.org/10.1002/mrm.28459
%
%   The demo data is acquired by SPICE sequence. To get the data, please 
%   send a request to rongguo2@illinois.edu and you will receive a link to 
%   directly download the data. 
%
%   If you also want to acquire SPICE data, the sequence file can be shared
%   once the C2P research agreement is done with University of illinois at
%   Urbana-Champaign. Current sequence is only implemented on Siemens
%   platform. 
%
%   The calculation of QSM is mainly based on the MEDI toolbox, to get the
%   latest version of MEDI toolbox, please refer to Cornell MRI Research
%   Lab: http://pre.weill.cornell.edu/mri/pages/qsm.html
%    
%   If any questions, please contact: rongguo2@illinois.edu
% ------------------------------------------------------------------------

% path setup 
codePath    = './code/';
dataPath    = './data/';
dataFile    = [dataPath,'demo_spice_data.mat'];
addpath(genpath(codePath));

% load prepared SPICE data 
varListSPICE = {...
    'ktSPICE',...               % k-t space SPICE data 
    'anatImag','dTime',...      % anatomical image and echospace 
    'watMask','watMaskSoft',... % brain mask and softmask
    'lipMask','lipMaskSoft',... % lipid mask and softmask
    'gapMask','gapMaskSoft',... % gap mask and softmask
    'indCenKy','indCenKz'...    % kspace index of central region
    };
load(dataFile,varListSPICE{:});

% basic parameter setup 
NyCen       = length(indCenKy);
NzCen       = length(indCenKz);

% GPS reconstruction on sparse data 
rankWat     = 12;
rankLip     = 20;
rankGap     = 15;
lambda      = 1e-2;
ktMaskGPS  	= abs(ktSPICE(:,:,:,:,1)) > 0;
optGPS      = struct('dt_D2',dTime,'L1_low',NyCen,'L3_low',NzCen,...
                     'watMask',watMask,'watMask_soft',watMaskSoft,...
                     'lipMask',lipMask,'lipMask_soft',lipMaskSoft,...
                     'gapMask',gapMask,'gapMask_soft',gapMaskSoft,...
                     'rank_W',rankWat,'rank_L',rankLip,'rank_G',rankGap,...
                     'enforCon',2,'verbose',2,'MaxIter',20,'lambda',lambda,...
                     'Iref',anatImag,'B0map',[],'flag_weights',true);
ktRecon     = gps_sparse_recon(ktSPICE,ktMaskGPS,optGPS);
ktRecon     = F3_k2x(ktRecon);
[sxtRecon1, senseMapHigh] = coilCombineSVD(ktRecon);

% removal of lipid signals 
lipRMmask   = bwmorph3D(imfill3D(lipMask,'holes'),'dilate',2)&(~watMask);
lipRMmask   = imresize3d(lipRMmask,size(ktRecon(:,:,:,1)));
ktRecon     = bsxfun(@times,ktRecon,lipRMmask);
ktRecon     = F3_x2k(ktRecon);
ktSPICE     = ktSPICE - ktRecon;
ktSPICE   	= bsxfun(@times,ktSPICE,ktMaskGPS);
clear ktRecon;

% second round GPS reconstruction 
[~,~,fmapHigh]    = hsvd3D(sxtRecon1,dTime,[-120,120],1,watMask);
goodMask    = applyMask3d(abs(fmapHigh)<10,watMask);
rankWat     = 20;
rankLip     = 1;
rankGap     = 1;
lambda      = 4e-3;
infoGPS     = struct('dt_D2',dTime,'L1_low',NyCen,'L3_low',NzCen,...
                     'watMask',watMask,'watMask_soft',watMaskSoft,...
                     'lipMask',lipMask,'lipMask_soft',lipMaskSoft,...
                     'gapMask',gapMask,'gapMask_soft',gapMaskSoft);
optGPS      = struct('rank_W',rankWat,'rank_L',rankLip,'rank_G',rankGap,...
                     'enforCon',1,'verbose',2,'MaxIter',20,'lambda',lambda,...
                     'Iref',anatImag,'B0map',fmapHigh,'senseMap',senseMapHigh,...
                     'goodMask',goodMask,'flag_weights',true);
ktRecon     = gps_sparse_recon(ktSPICE,ktMaskGPS,optGPS);
ktRecon     = F3_k2x(ktRecon);
sxtRecon2   = coilCombineSVD(ktRecon);
clear ktRecon ktSPICE;

% variables setup for QSM
varListQSM = {...
    'voxel_size',...        	% voxel size for QSM
    'matrix_size',...       	% matrix size for QSM 
    'delta_TE',...              % TE space for QSM
    'TE',...                    % TE vector for QSM
    'B0_dir',...                % B0 direction
    'CF'...                     % Central frequency in Hz
    };
load(dataFile,varListQSM{:});

iField      = imresize_ktrunc(sxtRecon2,matrix_size,0,0);
[Ny,Nx,Nz,Nt,Nc] = size(iField);
    
Mask_noise  = sqrt(squeeze(sum(sum(abs(iField).^2,4),5)));
for i=1:Nz
    Mask_noise(:,:,i) = get_mask(Mask_noise(:,:,i));
end
Mask_noise(:,:,[1:2,end-2:end]) = 1;

Mask        = imresize3d(watMask,matrix_size); 
Mask        = bwmorph3D(Mask,'erode',3);
ind_os      = round(0.2*matrix_size(3)/2);
Mask(:,:,[1:ind_os, end-ind_os:end]) = false;
noise_level = calfieldnoise(iField, Mask_noise);

iField      = iField/noise_level;
iMag        = anatImag;
for i=1:Nz
    Mask_noise(:,:,i) = get_mask(iMag(:,:,i));
end
noise_iMag  = calfieldnoise(iMag, Mask_noise);
iMag        = iMag/noise_iMag;

% QSM procesing
[~,~,fmap_total]    = hsvd3D(iField,delta_TE,[-inf,inf],1,Mask);
iFreq     	= fmap_total*2*pi*delta_TE; 
[iFreq_raw,N_std]   = Fit_ppm_linear(iField(:,:,:,1:32), matrix_size, voxel_size);
RDF        	= LBV(iFreq, Mask, matrix_size, voxel_size, 0.0001,2);
R2s         = arlo(TE,abs(iField));
Mask_CSF    = extract_CSF(R2s, Mask, voxel_size);
save('RDF.mat','RDF','iFreq','iMag','N_std','Mask','matrix_size','voxel_size','delta_TE','CF','B0_dir','Mask_CSF');
QSM         = MEDI_L1('lambda',2000,'merit','lambda_CSF',100);

% save variables 
varListSave = {...
    'sxtRecon2',...     	% reconstructed x-t data
    'QSM',...               % QSM map
    'RDF',...               % tissue field map
    'iFreq',...          	% total field map
    };
save(dataFile,'-append',varListSave{:});

% display
figure,montagesc(QSM(:,:,51:56));caxis([-0.12,0.12]);axis off;
